@@ -24,6 +24,7 @@
 #include <time.h>
 
 #include <netinet/in.h>
+#include <sys/param.h>
 #include <sys/socket.h>
 #include <sys/types.h>
 
@@ -52,7 +53,11 @@
 /* Minimal interval between subsequent sane_start()
  * attempts, if previous sane_start was failed
  */
-#define CONFIG_START_RETRY_INTERVAL      2500
+#define CONFIG_START_RETRY_INTERVAL     2500
+
+/* Default directory for AF_UNIX sockets
+ */
+#define CONFIG_DEFAULT_SOCKET_DIR       "/var/run"
 
 /******************** Forward declarations ********************/
 /* log_ctx represents logging context
@@ -570,6 +575,22 @@ __ptr_array_del (void **a, int i)
 #define safe_tolower(c)         tolower((unsigned char) c)
 
 /******************** OS Facilities ********************/
+/* The following macros, if defined, indicate that OS
+ * has a particular features:
+ *
+ *   OS_HAVE_EVENTFD   - Linux-like eventfd (2)
+ *   OS_HAVE_RTNETLINK - Linux-like rtnetlink (7)
+ *   OS_HAVE_AF_ROUTE  - BSD-like AF_ROUTE
+ */
+#ifdef  __linux__
+#   define OS_HAVE_EVENTFD   1
+#   define OS_HAVE_RTNETLINK 1
+#endif
+
+#ifdef BSD
+#   define OS_HAVE_AF_ROUTE  1
+#endif
+
 /* Get user's home directory. There is no need to
  * free the returned string
  *
@@ -792,6 +813,71 @@ uuid_equal (uuid u1, uuid u2)
     return !strcmp(u1.text, u2.text);
 }
 
+/******************** Generic .INI file parser ********************/
+/* Types of .INI file records
+ */
+typedef enum {
+    INIFILE_SECTION,                    /* The [section name] string */
+    INIFILE_VARIABLE,                   /* The variable = value string */
+    INIFILE_COMMAND,                    /* command param1 param2 ... */
+    INIFILE_SYNTAX                      /* The syntax error */
+} INIFILE_RECORD;
+
+/* .INI file record
+ */
+typedef struct {
+    INIFILE_RECORD      type;           /* Record type */
+    const char          *section;       /* Section name */
+    const char          *variable;      /* Variable name */
+    const char          *value;         /* Variable value */
+    const char          **tokv;         /* Value split to tokens */
+    unsigned int        tokc;           /* Count of strings in tokv */
+    const char          *file;          /* File name */
+    unsigned int        line;           /* File line */
+} inifile_record;
+
+/* .INI file (opaque)
+ */
+typedef struct {
+    const char          *file;                  /* File name */
+    unsigned int        line;                   /* File handle */
+    FILE                *fp;                    /* File pointer */
+
+    bool                tk_open;                /* Token is currently open */
+    char                *tk_buffer;             /* Parser buffer, tokenized */
+    unsigned int        *tk_offsets;            /* Tokens offsets */
+    unsigned int        tk_count;               /* Tokens count */
+
+    char                *buffer;                /* Parser buffer */
+    char                *section;               /* Section name string */
+    char                *variable;              /* Variable name string */
+    char                *value;                 /* Value string */
+    inifile_record      record;                 /* Record buffer */
+} inifile;
+
+/* Open the .INI file
+ */
+inifile*
+inifile_open (const char *name);
+
+/* Close the .INI file
+ */
+void
+inifile_close (inifile *file);
+
+/* Read next record
+ */
+const inifile_record*
+inifile_read (inifile *file);
+
+/* Match name of section of variable
+ *   - match is case-insensitive
+ *   - difference in amount of free space is ignored
+ *   - leading and trailing space is ignored
+ */
+bool
+inifile_match_name (const char *n1, const char *n2);
+
 /******************** Configuration file loader ********************/
 /* Device URI for manually disabled device
  */
@@ -826,6 +912,7 @@ typedef struct {
     bool        model_is_netname; /* Use network name instead of model */
     bool        proto_auto;       /* Auto protocol selection */
     WSDD_MODE   wsdd_mode;        /* WS-Discovery mode */
+    const char  *socket_dir;      /* Directory for AF_UNIX sockets */
 } conf_data;
 
 #define CONF_INIT {                     \
@@ -835,7 +922,8 @@ typedef struct {
         .discovery = true,              \
         .model_is_netname = true,       \
         .proto_auto = true,             \
-        .wsdd_mode = WSDD_FAST          \
+        .wsdd_mode = WSDD_FAST,         \
+        .socket_dir = NULL              \
     }
 
 extern conf_data conf;
@@ -856,7 +944,8 @@ conf_unload (void);
  * be passed by value
  */
 typedef struct {
-    char       text[64];
+    /* Holds sun_path from sockaddr_un plus a null byte. */
+    char       text[109];
 } ip_straddr;
 
 /* Format ip_straddr from IP address (struct in_addr or struct in6_addr)
@@ -866,7 +955,7 @@ ip_straddr
 ip_straddr_from_ip (int af, const void *addr);
 
 /* Format ip_straddr from struct sockaddr.
- * Both AF_INET and AF_INET6 are supported
+ * AF_INET, AF_INET6, and AF_UNIX are supported
  *
  * If `withzone' is true, zone suffix will be appended, when appropriate
  */
@@ -874,7 +963,7 @@ ip_straddr
 ip_straddr_from_sockaddr(const struct sockaddr *addr, bool withzone);
 
 /* Format ip_straddr from struct sockaddr.
- * Both AF_INET and AF_INET6 are supported
+ * AF_INET, AF_INET6, and AF_UNIX are supported
  *
  * Port will not be appended, if it matches provided default port
  * If `withzone' is true, zone suffix will be appended, when appropriate
@@ -963,6 +1052,11 @@ ip_addr_from_sockaddr (const struct sockaddr *sockaddr)
     return addr;
 }
 
+/* Format ip_addr into ip_straddr
+ */
+ip_straddr
+ip_addr_to_straddr (ip_addr addr, bool withzone);
+
 /* Check if two addresses are equal
  */
 static inline bool
@@ -1018,6 +1112,27 @@ ip_addrset_add_unsafe (ip_addrset *addrset, ip_addr addr);
 void
 ip_addrset_del (ip_addrset *addrset, ip_addr addr);
 
+/* Delete all addresses from the set
+ */
+void
+ip_addrset_purge (ip_addrset *addrset);
+
+/* Merge two sets:
+ *   addrset += addrset2
+ */
+void
+ip_addrset_merge (ip_addrset *addrset, const ip_addrset *addrset2);
+
+/* Get access to array of addresses in the set
+ */
+const ip_addr*
+ip_addrset_addresses (const ip_addrset *addrset, size_t *count);
+
+/* Check if two address sets are intersecting
+ */
+bool
+ip_addrset_is_intersect (const ip_addrset *set, const ip_addrset *set2);
+
 /******************** Network interfaces addresses ********************/
 /* Network interface name, wrapped into structure, so
  * it can be passed by value
@@ -1561,6 +1676,15 @@ http_query*
 http_query_new (http_client *client, http_uri *uri, const char *method,
         char *body, const char *content_type);
 
+/* Create new http_query
+ *
+ * Newly created http_query takes ownership on uri and body (if not NULL).
+ * The method and content_type assumed to be constant strings.
+ */
+http_query*
+http_query_new_len (http_client *client, http_uri *uri, const char *method,
+        void *body, size_t body_len, const char *content_type);
+
 /* Create new http_query, relative to base URI
  *
  * Newly created http_query takes ownership on body (if not NULL).
@@ -1753,7 +1877,8 @@ http_query_test_decode_response (http_query *q, const void *data, size_t size);
 typedef enum {
     HTTP_SCHEME_UNSET = -1,
     HTTP_SCHEME_HTTP,
-    HTTP_SCHEME_HTTPS
+    HTTP_SCHEME_HTTPS,
+    HTTP_SCHEME_UNIX
 } HTTP_SCHEME;
 
 /* Some HTTP status codes
@@ -2420,8 +2545,10 @@ typedef enum {
 typedef struct {
     ZEROCONF_METHOD   method;     /* Discovery method */
     const char        *name;      /* Network-unique name, NULL for WSD */
-    const char        *model;     /* Model name */
+    const char        *model;     /* Model name, may be NULL for
+                                     WSDD non-scanner devices */
     uuid              uuid;       /* Device UUID */
+    ip_addrset        *addrs;     /* Device addresses */
     int               ifindex;    /* Network interface index */
     zeroconf_endpoint *endpoints; /* List of endpoints */
 
@@ -2588,9 +2715,10 @@ device*
 device_open (const char *name, SANE_Status *status);
 
 /* Close the device
+ * If log_msg is not NULL, it is written to the device log as late as possible
  */
 void
-device_close (device *dev);
+device_close (device *dev, const char *log_msg);
 
 /* Get device's logging context
  */
@@ -3094,5 +3222,25 @@ log_panic (log_ctx *log, const char *fmt, ...);
 
 #endif
 
+/******************** Initialization/Cleanup ********************/
+/* AIRSCAN_INIT_FLAGS represents airscan_init() flags
+ */
+typedef enum {
+    AIRSCAN_INIT_NO_CONF        = (1 << 0),     // Don't load configuration
+    AIRSCAN_INIT_NO_THREAD      = (1 << 1)      // Don't start worker thread
+} AIRSCAN_INIT_FLAGS;
+
+/* Initialize airscan.
+ * If log_msg is not NULL, it is written to the log early
+ */
+SANE_Status
+airscan_init (AIRSCAN_INIT_FLAGS flags, const char *log_msg);
+
+/* Cleanup airscan
+ * If log_msg is not NULL, it is written to the log as late as possible
+ */
+void
+airscan_cleanup (const char *log_msg);
+
 /* vim:ts=8:sw=4:et
  */
